/*
#############################################################################

  Ch21p1_FiringRange.cpp: a program that demonstrates several different
  weaponry effects.
  
#############################################################################
*/

// include files ////////////////////////////////////////////////////////////
#define STRICT
#include <stdio.h>
#include <math.h>
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "D3DHelperFuncs.h"
#include "Ch21p1_resource.h"
#include "Camera.h"
#include "InputManager.h"
#include "SkyBox.h"
#include "Timer.h"
#include "GroundPlane.h"
#include "Ch21p1_Gun.h"
#include "Ch21p1_PlasmaGun.h"
#include "Ch21p1_LaserGun.h"
#include "Ch21p1_MachineGun.h"

const int NUMSPRITES = 5;

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
  // Font for drawing text
  CD3DFont* m_pFont;
  CD3DFont* m_pFontSmall;

  // Scene
  CUserControlledCamera m_Camera;

  // Mouse Input
  CInputManager m_InputManager;

  // skybox
  CSkyBox m_SkyBox;

	// ground plane
	CGroundPlane m_GroundPlane;

	CTimer m_TransitionTimer;

	// da weaponry
	CPlasmaGun m_PlasmaGun;
	CPlasmaBulletArray m_PlasmaBulletArray;

	CLaserGun m_LaserGun;
	CGun *m_pCurGun;
	CGun *m_pNextGun;

	CMachineGun m_MachineGun;
	
protected:
  HRESULT OneTimeSceneInit();
  HRESULT InitDeviceObjects();
  HRESULT RestoreDeviceObjects();
  HRESULT InvalidateDeviceObjects();
  HRESULT DeleteDeviceObjects();
  HRESULT FinalCleanup();
  HRESULT Render();
  HRESULT FrameMove();
  HRESULT ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format );
  LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

	void BeginGunTransition(CGun *pNextGun);

  void ProcessInput();
	int m_iLightPosX;
	int m_iLightPosY;

public:
  CMyD3DApplication();
};

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
  CMyD3DApplication d3dApp;

  if( FAILED( d3dApp.Create( hInst ) ) )
    return 0;

  return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
	m_strWindowTitle    = _T("Ch21p1_FiringRange");
  m_bUseDepthBuffer   = TRUE;
  m_pFont            = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
  m_pFontSmall       = new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
	if (m_fElapsedTime < 0.0001f) {
		m_fElapsedTime = 0.0001f;
	}
	
	CTimer::UpdateAll(m_fElapsedTime);
	m_PlasmaBulletArray.UpdateAll(m_fElapsedTime);
	if (m_pCurGun) m_pCurGun->Update(m_fElapsedTime);

	if (m_TransitionTimer.IsRunning()) {
		if (m_pCurGun) {
			if (m_TransitionTimer.GetTime() < 0.25f) {
			  m_pCurGun->Pos() = D3DXVECTOR3(0.5f, -0.6f-(m_TransitionTimer.GetTime()*4), 1.0f);
			}
			else {
				m_pCurGun = m_pNextGun;
				m_pCurGun->Pos() = D3DXVECTOR3(0.5f, -1.6f+(m_TransitionTimer.GetTime()*4)-1.0f, 1.0f);
			}
			if (m_TransitionTimer.GetTime() > 0.5f) {
				m_TransitionTimer.Stop();
				m_pNextGun = NULL;
			}
		}
	}
	return S_OK;
}

void CMyD3DApplication::BeginGunTransition(CGun *pNextGun)
{
	if (typeid(*pNextGun) == typeid(*m_pCurGun)) return;
	if (m_TransitionTimer.IsRunning()) return;

	m_pNextGun = pNextGun;
	m_TransitionTimer.Begin();
}

void CMyD3DApplication::ProcessInput()
{
  const float fSpeed = 0.5f;
  unsigned char m_bKey[256];
  ZeroMemory( m_bKey, 256 );
  GetKeyboardState(m_bKey);
  // Process keyboard input
  if(m_bKey['D'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(fSpeed, 0.0f, 0.0f)); // Slide Right
  if(m_bKey['A'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(-fSpeed, 0.0f, 0.0f));// Slide Left
  if(m_bKey['Q'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, fSpeed, 0.0f)); // Slide Up
  if(m_bKey['Z'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, -fSpeed, 0.0f));// Slide Down
  if(m_bKey['W'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, fSpeed)); // Slide Foward
  if(m_bKey['S'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, -fSpeed));// Slide Back
  if(m_bKey['L'] & 128) m_Camera.AddToYawPitchRoll(fSpeed, 0.0f, 0.0f);  // Turn Right
  if(m_bKey['J'] & 128) m_Camera.AddToYawPitchRoll(-fSpeed, 0.0f, 0.0f); // Turn Left
  if(m_bKey['K'] & 128) m_Camera.AddToYawPitchRoll(0.0f, fSpeed, 0.0f); // Turn Down
  if(m_bKey['I'] & 128) m_Camera.AddToYawPitchRoll(0.0f, -fSpeed, 0.0f);// Turn Up

  if(m_bKey['1'] & 128) BeginGunTransition(&m_MachineGun);
	if(m_bKey['2'] & 128) BeginGunTransition(&m_PlasmaGun);
	if(m_bKey['3'] & 128) BeginGunTransition(&m_LaserGun);

  // mouse look
  DIMOUSESTATE2 dims2;
  m_InputManager.ReadMouse(dims2);

  // play with the divisor constants to change the mouselook sensitivity.
  // I've found that these values most accurately simulate my beloved Q3A setup. :)
  m_Camera.AddToYawPitchRoll((float)dims2.lX/0.8f, (float)dims2.lY/0.8f, 0.0f);

	// prevent camera from moving up or down.
	D3DXVECTOR3 pos = m_Camera.GetPosition();
	pos.y = 2.60f; 
	m_Camera.SetPosition(pos);
  
	if (dims2.rgbButtons[0] && m_pCurGun && m_pCurGun->CanFire()) 
		m_pCurGun->Fire(m_Camera);
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
  
	ProcessInput();
  // these are done here so that you can move the camera around during a freeze
  // frame, ala The Matrix
  D3DXMATRIX matWorld;
  D3DXMatrixIdentity( &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

  ProcessInput();
  m_Camera.Update(m_fElapsedTime);

  m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
  m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
  m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
  
  // Clear the backbuffer
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                       0x000000, 1.0f, 0L );
  
  // Begin the scene  
  if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
  {
		m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1 );
		m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
		m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1 );
		m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
		
		m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
		m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

    // draw skybox
		m_pd3dDevice->SetVertexShader(D3DFVF_XYZ_DIFFUSE_TEX1);
		m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
		m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE);
		m_pd3dDevice->SetTransform( D3DTS_VIEW, &m_Camera.GetViewMatrix());
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE);
    m_SkyBox.Render(m_Camera.GetViewMatrix());

    
		// draw ground plane
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
		m_pd3dDevice->SetTransform( D3DTS_VIEW, &m_Camera.GetViewMatrix());
		m_pd3dDevice->SetTexture( 0, m_GroundPlane.GetTexture() );
    m_pd3dDevice->SetStreamSource( 0, m_GroundPlane.GetVB(), sizeof(VERTEX_XYZ_DIFFUSE_TEX1) );
    m_pd3dDevice->SetIndices( m_GroundPlane.GetIB(), 0L );
    m_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 
      m_GroundPlane.GetNumVerts(), 0, m_GroundPlane.GetNumIndices()/3 );

		// draw gun
		m_pCurGun->Render(m_Camera);

		// draw plasma bullets
		m_PlasmaBulletArray.RenderAll(m_Camera);
		
    char buf[256];
    _snprintf(buf, 256, "Position: (%.2f, %.2f, %.2f)", 
      (float)m_Camera.GetPosition().x, (float)m_Camera.GetPosition().y, 
      (float)m_Camera.GetPosition().z);

    m_pFontSmall->DrawText( 2,  0,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    // End the scene.
    m_pd3dDevice->EndScene();
  }

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
  m_pFont->InitDeviceObjects( m_pd3dDevice );
  m_pFontSmall->InitDeviceObjects( m_pd3dDevice );
	m_SkyBox.SetSize(100.0f);
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
	m_iLightPosX = 0;
	m_iLightPosY = 0;

	m_InputManager.CreateDevices(m_hWnd, true, true);

  m_SkyBox.RestoreDeviceObjects(m_pd3dDevice,
    "Ch21p1_DesertSkyboxTop.bmp",
    "Ch21p1_DesertSkyboxBottom.bmp",
    "Ch21p1_DesertSkyboxFront.bmp",
    "Ch21p1_DesertSkyboxBack.bmp",
    "Ch21p1_DesertSkyboxLeft.bmp",
    "Ch21p1_DesertSkyboxRight.bmp"
    );

	m_GroundPlane.RestoreDeviceObjects(m_pd3dDevice, "Ch21p1_GroundTexture.bmp", 512.0f, 512.0f, 16);

	m_PlasmaGun.RestoreDeviceObjects(m_pd3dDevice, "Ch21p1_Gun.x");
	m_PlasmaBulletArray.RestoreDeviceObjects(m_pd3dDevice);
	m_PlasmaGun.SetBulletArray(&m_PlasmaBulletArray);

	m_LaserGun.RestoreDeviceObjects(m_pd3dDevice, "Ch21p1_Gun.x");
	m_MachineGun.RestoreDeviceObjects(m_pd3dDevice, "Ch21p1_Gun.x");

	m_PlasmaGun.Pos() = D3DXVECTOR3(0.5f, -0.6f, 1.0f);
	m_LaserGun.Pos() = D3DXVECTOR3(0.5f, -0.6f, 1.0f);
	m_MachineGun.Pos() = D3DXVECTOR3(0.5f, -0.6f, 1.0f);

	m_pCurGun = &m_MachineGun;

	m_pFont->RestoreDeviceObjects();
  m_pFontSmall->RestoreDeviceObjects();
  m_Camera.SetPosition(D3DXVECTOR3(0.0f, 7.0f, -20.0f));
	
	// Set the world matrix
  D3DXMATRIX matWorld;
  D3DXMatrixIdentity( &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

  // Set projection matrix
  D3DXMATRIX matProj;
  FLOAT fAspect = ((FLOAT)m_d3dsdBackBuffer.Width) / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, fAspect, 0.1f, 100.0f );
  m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE );
  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

  m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,  FALSE );
  m_pd3dDevice->SetRenderState( D3DRS_CULLMODE,  D3DCULL_CCW );
  m_pd3dDevice->SetRenderState( D3DRS_SHADEMODE, D3DSHADE_GOURAUD );
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
  m_InputManager.DestroyDevices();
  m_pFont->InvalidateDeviceObjects();
  m_pFontSmall->InvalidateDeviceObjects();
  m_SkyBox.InvalidateDeviceObjects();
	m_GroundPlane.InvalidateDeviceObjects();
	
	m_PlasmaGun.InvalidateDeviceObjects();
	m_PlasmaBulletArray.InvalidateDeviceObjects();

	m_LaserGun.InvalidateDeviceObjects();
	m_MachineGun.InvalidateDeviceObjects();

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
  m_pFont->DeleteDeviceObjects();
  m_pFontSmall->DeleteDeviceObjects();
	m_GroundPlane.DeleteDeviceObjects();
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
  SAFE_DELETE( m_pFont );
  SAFE_DELETE( m_pFontSmall );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam )
{
  // Pass remaining messages to default handler
  return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}